@@ -2,8 +2,8 @@ import erlang
 import requests
 import json
 
-from . import DEFAULT_API_URL, logger, ArweaveException
-from .utils import response_stream_to_file_object, b64dec, be_u256dec
+from . import DEFAULT_API_URL, logger, ArweaveException, ArweaveNetworkException
+from .utils import response_stream_to_file_object, b64dec, arbindec
 
 class HTTPClient:
     def __init__(self, api_url, timeout = None, retries = 5):
@@ -22,11 +22,16 @@ class HTTPClient:
 
         response = self.session.request(**{'method': 'GET', 'url': url, 'timeout': self.timeout, **request_kwparams})
 
-        if response.status_code >= 200 and response.status_code < 300 and int(response.headers.get('content-length', 1)) > 0:
+        try:
+            response.raise_for_status()
+            if int(response.headers.get('content-length', 1)) == 0:
+                raise ArweaveException('Empty response.')
+            if response.status_code not in (200, 206):
+                raise ArweaveException(response.text)
             return response
-        else:
+        except requests.exceptions.RequestException as exc:
             logger.error(response.text)
-            raise ArweaveException(response.text)
+            raise ArweaveNetworkException(response.text, exc, response)
 
     def _post(self, data, *params, headers = {}, **request_kwparams):
         if len(params) and params[-1][0] == '?':
@@ -49,12 +54,13 @@ class HTTPClient:
 
         # logger.debug('{}\n\n{}'.format(response.text, data))
 
-        if response.status_code >= 200 and response.status_code < 300:
+        try:
+            response.raise_for_status()
             # logger.debug('RESPONSE 200: {}'.format(response.text))
             return response
-        else:
-            # logger.error('{}\n\n{}'.format(response.text, data))
-            raise ArweaveException(response.text, data, url)
+        except requests.exceptions.RequestException as exc:
+            logger.error('{}\n\n{}'.format(response.text, data))
+            raise ArweaveNetworkException(response.text, exc, response)
 
 class Peer(HTTPClient):
     # peer api [incomplete]:
@@ -84,6 +90,108 @@ class Peer(HTTPClient):
         response = self._get('info')
         return response.json()
 
+    def probe(self):
+        features = []
+        systems = []
+        status = {}
+        exclude = set()
+
+        def load_balancing_proxy():
+            sync_record_1 = self.data_sync_record(0,1)[-1]
+            sync_record_2 = self.data_sync_record(0,2)[-1]
+            if sync_record_1 == sync_record_2:
+                raise ArweaveNetworkException()
+
+        def partial_data():
+            for height in range(self.height(), 0, -1):
+                tx = self.block_height(height)['txs'][0]
+                try:
+                    stream = self.stream(tx, range = (0,4))
+                    data = stream.read(8)
+                    if len(data) == 8:
+                        partial_data = False
+                    else:
+                        stream = self.stream(tx)
+                        data = stream.read(8)
+                        if len(data) == 8:
+                            partial_data = True
+                        else:
+                            continue
+                    break
+                except ArweaveException as exc:
+                    print(tx, exc)
+                    continue
+            if not partial_data:
+                raise ArweaveNetworkException()
+
+        def peers():
+            health = status.get('health')
+            if health is not None:
+                status['peers'] = [origin['endpoint'].split('://',1)[1] for origin in health['origins']]
+                raise ArweaveNetworkException()
+            else:
+                return self.peers()
+
+        probes = dict(
+            arql = dict(
+                params = [dict(op='equals',expr1='',expr2='')],
+            ),
+            graphql = dict(
+                params = ['query{transaction(id:""){id}}']
+            ),
+            load_balancing_proxy = dict(
+                probe = load_balancing_proxy,
+                exclude = ['info', 'sync_buckets']
+            ),
+            partial_data = dict(
+                probe = partial_data,
+            ),
+            cache_jobs = dict(
+                system = 'arseeding',
+            ),
+            health = dict(
+                system = 'arweave-gateway',
+            ),
+            info = dict(
+                system = 'arweave-erlang'
+            ),
+            peers = dict(
+                probe = peers,
+                system = 'arweave-erlang'
+            ),
+            sync_buckets = dict(
+                system = 'arweave-erlang'
+            )
+        )
+
+        for probe, options in probes.items():
+            if probe in exclude:
+                continue
+            logger.info(f'{probe} ...')
+            try:
+                params = options.get('params',[])
+                func = options.get('probe')
+                if func is None:
+                    func = getattr(self, probe)
+                reply = func(*options.get('params',[]))
+                logger.info(f'{self.api_url} has {probe}')
+                if len(params) == 0 and reply is not None:
+                    status[options.get('status',probe)] = reply
+                exclude.update(options.get('exclude',[]))
+                features.append(options.get('feature',probe))
+                system = options.get('system')
+                if system is not None and system not in systems:
+                    systems.append(system)
+            except ArweaveNetworkException:
+                logger.info(f'{self.api_url} does not have {probe}')
+                continue
+ 
+        return {
+            'guess': systems,
+            'features': features,
+            'status': status
+        }
+
     def time(self):
         '''Return the current universal time in seconds.'''
         response = self._get('time')
@@ -178,14 +286,6 @@ class Peer(HTTPClient):
         response = self._post(logical_expression, 'arql')
         return response.json()
 
-    def graphql(self, query):
-        response = self._post({
-            'operationName': None,
-            'query': query,
-            'variables': {}
-        }, 'graphql')
-        return response.json()
-
     def tx_data_html(self, txid):
         '''
         Return the data field of the transaction specified via the transaction ID (hash)
@@ -237,13 +337,15 @@ class Peer(HTTPClient):
                 pass
         intervals = erlang.binary_to_term(response.content)
         intervals = [
-            (be_u256dec(left.value), be_u256dec(right.value))
+            (int.from_bytes(left.value, 32, 'big'), int.from_bytes(right.value, 32, 'big'))
             for left, right in intervals
         ]
         return intervals
 
     def chunk(self, offset, packing = 'unpacked', bucket_based_offset = False):
         '''
+        Returns the data chunk containing 1-based offset,
+        using json, base64 encoded network transmission.
 
         {packing} := { 'unpacked' | 'spora_2_5' | 'any' }
 
@@ -269,10 +371,12 @@ class Peer(HTTPClient):
 
     def chunk2(self, offset, packing = 'unpacked', bucket_based_offset = False):
         '''
+        Returns the data chunk containing 1-based offset,
+        using raw binary network transmission.
 
         {packing} := { 'unpacked' | 'spora_2_5' | 'any' }
 
-        Returns: b[
+        The raw data format from the endpoint is b[
             chunk_size      3 bytes, big-endian
             chunk           chunk_size bytes
 
@@ -285,6 +389,14 @@ class Peer(HTTPClient):
             packing2_size   1 byte
             packing2        packing2_size bytes
         ]
+
+        This function parses it to return the same format as chunk():
+        {
+            "chunk": b"<chunk>",
+            "tx_path": b"<txpath>",
+            "data_path"; b"<datapath>",
+            "packing": "packing"
+        }
         '''
 
         headers = {
@@ -293,21 +405,44 @@ class Peer(HTTPClient):
         if bucket_based_offset:
             headers['x-bucket-based-offset'] = '1'
 
-        response = self._get('chunk2', str(offset), headers=headers)
-        return response.content
+        response = self._get('chunk2', str(offset), headers=headers, stream = True)
+        with response_stream_to_file_object(response) as stream:
+            result = {}
+            result['chunk'] = arbindec(stream, 24)
+            result['tx_path'] = arbindec(stream, 24)
+            result['data_path'] = arbindec(stream, 24)
+            result['packing'] = arbindec(stream, 8).decode()
+            extra = stream.read()
+            if extra:
+                result['_extra'] = extra
+            return result
+
+    def chunk_size(self, offset, packing = 'unpacked', bucket_based_offset = False):
+        '''Returns the size of the data chunk containing 1-based offset.'''
+
+        headers = {
+            'x-packing': packing,
+            'Range': 'bytes=0-2',
+        }
+        if bucket_based_offset:
+            headers['x-bucket-based-offset'] = '1'
+
+        response = self._get('chunk2', str(offset), headers=headers, stream = True)
+        with response_stream_to_file_object(response) as stream:
+            return int.from_bytes(stream.read(3), 'big')
 
     def tx_offset(self, hash):
         '''
         Get the absolute end offset and size of the transaction
 
-        The client may use this information to collect transaction chunks. Start with
-        the end offset and fetch a chunk via chunk(<offset>). Subtract its size
-        from the transaction size - if there are more chunks to fetch, subtract the
-        size of the chunk from the offset and fetch the next chunk.
+        The client may use this information to collect transaction chunks. Add 1 to
+        the end offset and substract the size to get the start offset, then fetch a
+        chunk via chunk(<start offset>). Add its size to the start offset and fetch
+        the next chunk - if there are more chunks, continue to do the same.
 
         {
-            "offset": <Offset>,
-            "size": <Size>
+            "offset": <offset of last byte of tx data>,
+            "size": <total size of tx data>
         }
         '''
         response = self._get('tx', hash, 'offset')
@@ -745,3 +880,87 @@ class Peer(HTTPClient):
             headers = {}
         response = self._get(txid + ext, headers = headers, stream = True)
         return response_stream_to_file_object(response)
+
+    # below are used with https://github.com/ar-io/arweave-gateway
+
+    def graphql(self, query):
+        response = self._post({
+            'operationName': None,
+            'query': query,
+            'variables': {}
+        }, 'graphql')
+        return response.json()
+
+    def health(self):
+        '''
+        Returns information on the connected peers and database.
+        
+        {
+            "region": "<AWS_REGION>",
+            "origins": [
+                {
+                    "endpoint": "<api_url>",
+                    "status": <http status code>, 
+                    "info": peer.info()
+                }
+            ],
+            "database": {
+                "block": { "id", "height", "mined_at", "previous_block", "txs", "extended" }
+            }
+        }
+        '''
+        response = self._get('health')
+        return response.json()
+        
+
+    # below are used with https://github.com/everFinance/arseeding
+
+    def send_job_broadcast(self, txid):
+        '''Register a tx to be broadcast to all nodes.'''
+        response = self._get('job', 'broadcast', txid, method='POST')
+        return response.text
+
+    def send_job_sync(self, txid):
+        response = self._get('job', 'sync', txid, method='POST')
+        return response.text
+
+    def job_kill(self, txid, type):
+        '''Close a running job.'''
+        response = self._get('job', 'kill', txid, type, method='POST')
+        return response.text
+
+    def job_kill_broadcast(self, txid):
+        '''Close a broadcast job.'''
+        return self.job_kill(txid, 'broadcast')
+
+    def job_kill_sync(self, txid):
+        '''Close a sync job.'''
+        return self.job_kill(txid, 'sync')
+
+    def job(self, txid, type):
+        '''Get the status of a job.'''
+        response = self._get('job', txid, type)
+        return response.json()
+
+    def job_broadcast(self, txid):
+        '''Get the status of a broadcast job.
+
+        {
+            "arid": "<txid>",
+            "jobType": "broadcast",
+            "countSuccessed": <nodes successfully broadcast>,
+            "countFailed": <nodes failed to broadcast>,
+            "totalNodes": <total number of nodes>,
+            "close": <whether task is closed>
+        }
+        '''
+        return self.job(txid, 'broadcast')
+
+    def job_sync(self, txid):
+        '''Get the status of a sync job.
+        '''
+        return self.job(txid, 'sync')
+
+    def cache_jobs(self):
+        response = self._get('cache', 'jobs')
+        return response.json()
